#include <gtk/gtk.h>
#include <gtkmozembed.h>
#include "light.h"

#include "ui.h"
#include "context.h"
#include "fifo.h"
#include "moz_main.h"
#include "utils.h"

/* kinda hackish way to keep track of the selected menu item */
static int context_current_item;

/* obviously, the list below must match this enum */
typedef enum {
    CTX_OPEN = 0,			/* open link in new window */
    CTX_OPEN_FRAME,			/* open frame in new window */
    CTX_OPEN_IMG,			/* open image in new window */

    /* clipboard */
    CTX_COPY_LINK,			/* copy link location */
    CTX_COPY_FRAME,			/* copy frame location */
    CTX_COPY_IMG,			/* copy img location */
    
    /* bookmarks / fifo */
    CTX_FIFO_LINK,			/* copy link to fifo */
    CTX_FIFO_FRAME,			/* copy frame to fifo */
    CTX_FIFO_IMG,			/* copy img to fifo */
    CTX_BM_LINK,			/* copy link to bookmarks */
    CTX_BM_FRAME,			/* copy frame to bookmarks */

    /* downloading */
    CTX_DOWN_LINK,			/* download link */
    CTX_SAVE_IMG,			/* save image */

    CTX_SEPARATOR,			/* separator */
    CTX_END				/* end of context menu */
} ContextType;

static char *context_items[] = {
    "Open link in new window",
    "Open frame in new window",
    "Open image in new window",

    "Copy link location to Clipboard",
    "Copy frame location to Clipboard",
    "Copy image location to Clipboard",

    "Copy link location to Fifo",
    "Copy frame location to Fifo",
    "Copy image location to Fifo",
    "Add link to bookmarks",
    "Add frame to bookmarks",

    "Download link",
    "Save image"
};

/* popup menus handler */
static int context_menu_popup_modal(GtkWidget *menu, unsigned int timestamp);
static void context_menu_activated(GtkWidget *item, gpointer data);
static void context_menu_deactivated(GtkWidget *item, gpointer data);
static GtkWidget *context_menu_new(int dummy, ...);

static void context_run(Browser *browser, ContextType type, MouseEventInfo *info);
static void light_save_as_cb(GtkWidget *button, GtkWidget *dialog);

/**
 * context_popup_menu: open up a context menu at current mouse location,
 * wait for user input, and run appropriate callback
 */
void context_popup_menu(Browser *browser, MouseEventInfo *info)
{
    GtkWidget *menu = NULL;
    int context;
    int result;
    
    context = info->context;

    if ((context & CONTEXT_IMAGE) && (context & CONTEXT_LINK)) {
	/* image + link */
	menu = context_menu_new(TRUE,
		CTX_OPEN,
		CTX_DOWN_LINK,
		CTX_OPEN_IMG,
		CTX_SAVE_IMG,
		CTX_SEPARATOR,
		CTX_COPY_LINK,
		CTX_FIFO_LINK,
		CTX_COPY_IMG,
		CTX_FIFO_IMG,
		CTX_BM_LINK,
		CTX_END);
    } else if (context & CONTEXT_IMAGE) {
	/* image */
	menu = context_menu_new(TRUE,
		CTX_OPEN_IMG,
		CTX_SAVE_IMG,
		CTX_SEPARATOR,
		CTX_COPY_IMG,
		CTX_FIFO_IMG,
		CTX_END);
    } else if (context & CONTEXT_LINK) {
	/* link */
	menu = context_menu_new(TRUE,
		CTX_OPEN,
		CTX_DOWN_LINK,
		CTX_SEPARATOR,
		CTX_COPY_LINK,
		CTX_FIFO_LINK,
		CTX_BM_LINK,
		CTX_END);
    } else if (context & CONTEXT_DOCUMENT && info->framed) {
	/* frame */
	menu = context_menu_new(TRUE,
		CTX_OPEN_FRAME,
		CTX_SEPARATOR,
		CTX_COPY_FRAME,
		CTX_FIFO_FRAME,
		CTX_BM_FRAME,
		CTX_END);
    }

    if (menu) {
	/* oh, that was a fucking GTK bug */
	result = context_menu_popup_modal(menu, info->timestamp);
	context_run(browser, result, info);
    }
}

void light_save_as(Browser *browser, const char *url)
{
    char *filename;
    GtkWidget *dialog;

    filename = url_basename(url);

    dialog = gtk_file_selection_new("Save As...");
    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(dialog));
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(dialog), filename);
    gtk_object_set_data_full(GTK_OBJECT(dialog), "url", g_strdup(url), g_free);
    gtk_object_set_data(GTK_OBJECT(dialog), "browser", browser);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->ok_button),
		       "clicked", GTK_SIGNAL_FUNC(light_save_as_cb),
		       dialog);
    gtk_signal_connect_object(GTK_OBJECT
			      (GTK_FILE_SELECTION(dialog)->cancel_button),
			      "clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      (gpointer) dialog);
    gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
		       GTK_SIGNAL_FUNC(dialog_esc_handler), dialog);
    gtk_widget_show_all(dialog);
    g_free(filename);
}

static void context_run(Browser *browser, ContextType type, MouseEventInfo *info)
{
    char *url = NULL;

    switch (type) {
	case CTX_OPEN:
	    light_window_new(info->link);
	    break;
	case CTX_OPEN_FRAME:
	    url = moz_get_document_url(browser);
	    if (url) {
		light_window_new(url);
		g_free(url);
	    }
	    break;
	case CTX_OPEN_IMG:
	    light_window_new(info->img);
	    break;
	case CTX_COPY_LINK:
	    light_copy_clipboard(browser, info->link);
	    break;
	case CTX_COPY_FRAME:
	    url = moz_get_document_url(browser);
	    if (url) {
		light_copy_clipboard(browser, url);
		g_free(url);
	    }
	    break;
	case CTX_COPY_IMG:
	    light_copy_clipboard(browser, info->img);
	    break;
	case CTX_FIFO_LINK:
	    light_add_fifo(browser, info->link, info->linktext);
	    break;
	case CTX_FIFO_FRAME:
	    url = moz_get_document_url(browser);
	    if (url) {
		light_add_fifo(browser, url, NULL);
		g_free(url);
	    }
	    break;
	case CTX_FIFO_IMG:
	    light_add_fifo(browser, info->img, NULL);
	    break;
	case CTX_BM_LINK:
	    // context_add_bookmark(browser, info->link, info->linktext);
	    break;
	case CTX_BM_FRAME:
	    url = moz_get_document_url(browser);
	    if (url) {
		// context_add_bookmark(browser, url, NULL);
		g_free(url);
	    }
	    break;
	case CTX_DOWN_LINK:
	    light_download(browser, info->link);
	    break;
	case CTX_SAVE_IMG:
	    light_save_as(browser, info->img);
	    break;
	case CTX_SEPARATOR:
	case CTX_END:
	    break;
    }
}

/**
 * context_menu_popup_modal: popup a context menu and wait until an item
 * is selected
 */
static int context_menu_popup_modal(GtkWidget *menu, unsigned int timestamp)
{
    context_current_item = CTX_END;

    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
	    3, timestamp);

    gtk_grab_add(menu);
    gtk_main();
    gtk_grab_remove(menu);
    gtk_widget_destroy(menu);
    return context_current_item;
}

/**
 * context_menu_activated: called by the selected menu item to set the
 * global context_current_item
 */
static void context_menu_activated(GtkWidget *item, gpointer data)
{
    context_current_item = GPOINTER_TO_INT(data);
}

/**
 * context_menu_deactivated: called when the menu is unmapped - to remove
 * the gtk grab
 */
static void context_menu_deactivated(GtkWidget *item, gpointer data)
{
    gtk_main_quit();
}

/**
 * context_menu_new: make a new context menu
 * @dummy: bogus variable
 * @...: list of ContextType menu items, terminated with CTX_END
 */
static GtkWidget *context_menu_new(int dummy, ...)
{
    GtkWidget *menu;
    GtkWidget *menuitem;

    /*  menu item variables  */
    ContextType type;

    va_list args;
    int i;

    menu = gtk_menu_new();
    gtk_container_set_border_width(GTK_CONTAINER(menu), 1);

    va_start(args, dummy);

    type = va_arg(args, ContextType);
    for (i = 0; type != CTX_END; i++) {
	if (type == CTX_SEPARATOR) {
	    /* make a separator */
	    menuitem = gtk_menu_item_new();
	    gtk_widget_set_sensitive(menuitem, FALSE);
	} else {
	    /* make a normal menu item */
	    menuitem = gtk_menu_item_new_with_label(context_items[type]);
	    gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
			       context_menu_activated, GINT_TO_POINTER(type));
	    if ((type == CTX_BM_LINK) || (type == CTX_BM_FRAME))
		gtk_widget_set_sensitive(menuitem, FALSE);
	}

	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_widget_show(menuitem);

	type = va_arg(args, ContextType);
    }
    va_end(args);

    gtk_signal_connect(GTK_OBJECT(menu), "deactivate",
	    (GtkSignalFunc) context_menu_deactivated, NULL);

    return menu;
}

static void light_save_as_cb(GtkWidget *button, GtkWidget *dialog)
{
    char *filename;
    char *url;
    Browser *browser;

    filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
    url = gtk_object_get_data(GTK_OBJECT(dialog), "url");
    browser = gtk_object_get_data(GTK_OBJECT(dialog), "browser");

    if (!filename || !strcmp(filename, ""))
	return;

    moz_save_cached(browser, url, filename);
    gtk_widget_destroy(dialog);
}

