#include <gtk/gtk.h>
#include <string.h>
#include <gtkmozembed.h>
#include "light.h"

#include "context.h"
#include "moz_cb.h"
#include "moz_main.h"
#include "moz_pref.h"
#include "ui.h"
#include "utils.h"

extern int win_count;
extern GList *win_list;
extern Config config;

/* local functions */
static void update_buttons(Browser *browser);
static int throbber_loop(Browser *browser);

static void embed_destroy_cb(GtkWidget *widget, Browser *browser);
static void embed_destroy_brsr_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_new_window_cb(GtkMozEmbed *embed, GtkMozEmbed **new, unsigned int chrome, Browser *browser);
static void embed_visibility_cb(GtkMozEmbed *embed, int vis, Browser *browser);
static void embed_size_cb(GtkMozEmbed *embed, int width, int height, Browser *browser);
static void embed_location_changed_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_title_changed_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_link_message_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_js_status_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_load_started_cb(GtkMozEmbed *embed, Browser *browser);
static void embed_load_finished_cb(GtkMozEmbed *embed, Browser *browser);
static  int embed_open_uri_cb(GtkMozEmbed *embed, const char *uri, Browser *browser);
static void net_state_change_cb(GtkMozEmbed *embed, int flags, unsigned int status, Browser *browser);
static void progress_change_cb(GtkMozEmbed *embed, int cur, int max, Browser *browser);

/* dom events */
static int dom_key_press_cb(GtkMozEmbed *embed, void *event, Browser *browser);
static int dom_mouse_down_cb(GtkMozEmbed *embed, void *event, Browser *browser);
static int dom_mouse_click_cb(GtkMozEmbed *embed, void *event, Browser *browser);

/* orphan window junk */
static void embed_new_orphan_window_cb(GtkMozEmbedSingle *embed, GtkMozEmbed **retval, guint chromemask, gpointer data);

static const struct {
    char *event;
    void *func;
} signal_connections[] = {
#if 0
    /* these guys are not really necessary until I know how to use them */
    { "net_state_all", net_state_change_all_cb },
    { "progress_all", progress_change_all_cb },
    { "dom_mouse_up", dom_mouse_up_cb },
    { "dom_mouse_dbl_click", dom_mouse_dbl_click_cb },
    { "dom_mouse_over", dom_mouse_over_cb },
    { "dom_mouse_out", dom_mouse_out_cb },
#endif

    /* keyboard / mouse handlers */
    { "dom_mouse_down", dom_mouse_down_cb },
    { "dom_mouse_click", dom_mouse_click_cb },
    { "dom_key_press", dom_key_press_cb },

    /* browser stuff */
    { "progress", progress_change_cb },
    { "net_state", net_state_change_cb },
    { "open_uri", embed_open_uri_cb },
    { "net_start", embed_load_started_cb },
    { "net_stop", embed_load_finished_cb },
    { "link_message", embed_link_message_cb },
    { "js_status", embed_js_status_cb },
    { "title", embed_title_changed_cb },
    { "location", embed_location_changed_cb },
    { "size_to", embed_size_cb },
    { "visibility", embed_visibility_cb },
    { "new_window", embed_new_window_cb },
    { "destroy_browser", embed_destroy_brsr_cb },
    { "destroy", embed_destroy_cb },
    { NULL, NULL }
};

static int throbber_loop(Browser *browser)
{
    char *spinner[] = { ".", "o", "O", "o" };

    if (!browser->sb.label) {
	gtk_timeout_remove(browser->load.throbber);
	browser->load.throbber = 0;
	return FALSE;
    }

    gtk_label_set_text(GTK_LABEL(browser->sb.label), spinner[browser->load.spin]);
    browser->load.spin++;
    if (browser->load.spin > 3)
	browser->load.spin = 0;
    return TRUE;
}

void moz_connect_signals(Browser *browser)
{
    int i;

    for (i = 0; signal_connections[i].event != NULL; i++) {
	gtk_signal_connect(GTK_OBJECT(browser->embed),
			   signal_connections[i].event,
			   signal_connections[i].func, browser);
    }
}

/* update toolbar buttons */
static void update_buttons(Browser *browser)
{
    if (gtk_moz_embed_can_go_back(GTK_MOZ_EMBED(browser->embed)))
	gtk_widget_set_sensitive(browser->btn.back, TRUE);
    else
	gtk_widget_set_sensitive(browser->btn.back, FALSE);

    if (gtk_moz_embed_can_go_forward(GTK_MOZ_EMBED(browser->embed)))
	gtk_widget_set_sensitive(browser->btn.forward, TRUE);
    else
	gtk_widget_set_sensitive(browser->btn.forward, FALSE);
}

/* mozilla widget callbacks */
static void embed_destroy_cb(GtkWidget *widget, Browser *browser)
{
    GList *tmp;

    win_count--;
    tmp = g_list_find(win_list, browser);
    win_list = g_list_remove_link(win_list, tmp);
   
    /* get rid of some temp stuff */
    if (browser->sb.link)
	g_free(browser->sb.link);
    if (browser->clipboard)
	g_free(browser->clipboard);
    if (browser->find)
	g_free(browser->find);

    /* nuke menubar factory */
    gtk_object_destroy(GTK_OBJECT(browser->fact));

    g_free(browser);
    g_list_free(tmp);
    browser = NULL;

    if (win_count == 0)
	gtk_main_quit();
}

static void embed_new_window_cb(GtkMozEmbed *embed,
	GtkMozEmbed **new, unsigned int chrome, Browser *browser)
{
    Browser *new_browser = NULL;

    new_browser = light_new(chrome);
    *new = GTK_MOZ_EMBED(new_browser->embed);
}

static void embed_destroy_brsr_cb(GtkMozEmbed *embed, Browser *browser)
{
    gtk_widget_destroy(browser->window);
}

static void embed_visibility_cb(GtkMozEmbed *embed, int vis,
	Browser *browser)
{
    light_set_visibility(browser, vis);
}

static void embed_size_cb(GtkMozEmbed *embed, int width, int height,
	Browser *browser)
{
    int mheight = 0;
    int theight = 0;
    int sheight = 0;
    int max = 0;
    unsigned int chrome;
    
    /* resize browser widget to required size */
    gtk_widget_set_usize(browser->embed, width, height);

    /* get chrome stuff */
    chrome = gtk_moz_embed_get_chrome_mask(GTK_MOZ_EMBED(browser->embed));
    if (chrome == GTK_MOZ_EMBED_FLAG_DEFAULTCHROME)
	chrome = GTK_MOZ_EMBED_FLAG_ALLCHROME;

    /* see if we need to adjust size for missing elements */
    if (chrome & GTK_MOZ_EMBED_FLAG_MENUBARON) {
	mheight = browser->menubar->allocation.height;
	max++;
    }
    if (chrome & GTK_MOZ_EMBED_FLAG_TOOLBARON) {
	theight = browser->toolbar->allocation.height;
	max++;
    }
    if (chrome & GTK_MOZ_EMBED_FLAG_STATUSBARON) {
	sheight = browser->sb.bar->allocation.height;
	max++;
    }

    /* account for the menubar + toolbar */
    height += (theight + mheight + (UI_SPACING * max) + sheight);
    
    gtk_widget_set_usize(browser->window, width, height);
    gtk_window_set_default_size(GTK_WINDOW(browser->window),
	    width, height);
}

static void embed_location_changed_cb(GtkMozEmbed *embed, Browser *browser)
{
    char *new;
    int pos = 0;

    new = gtk_moz_embed_get_location(embed);
    if (new) {
	gtk_editable_delete_text(GTK_EDITABLE(browser->entry), 0, -1);
	gtk_editable_insert_text(GTK_EDITABLE(browser->entry),
		new, strlen(new), &pos);
	g_free(new);
    }
    update_message(browser, 0);
    update_buttons(browser);
}

static void embed_title_changed_cb(GtkMozEmbed *embed, Browser *browser)
{
    char *title;

    /* title = gtk_moz_embed_get_title(embed); */
    title = moz_get_document_title(browser);
    if (title) {
	char *new = NULL;
	if (!strlen(title))
	    new = g_strdup("Light");
	else
	    new = g_strconcat("Light - ", title, NULL);
	gtk_window_set_title(GTK_WINDOW(browser->window), new);
	g_free(new);
	g_free(title);
    }
}

static void embed_link_message_cb(GtkMozEmbed *embed, Browser *browser)
{
    char *message;

    message = gtk_moz_embed_get_link_message(embed);
    if (message && (strlen(message) == 0))
	update_message(browser, 0);
    else
	update_message(browser, message);
    if (message)
	g_free(message);
}

static void embed_js_status_cb(GtkMozEmbed *embed, Browser *browser)
{
    char *message;

    message = gtk_moz_embed_get_js_status(embed);
    if (message && (strlen(message) == 0))
	update_message(browser, 0);
    else
	update_message(browser, message);

    if (message)
	g_free(message);
}

static void embed_load_started_cb(GtkMozEmbed *embed, Browser *browser)
{
    browser->load.st_time = time(NULL);
    browser->load.percent = 0;
    browser->load.bytes = 0;
    browser->load.maxbytes = 0;

    /* activate throbber */
    if (!browser->load.throbber) {
	browser->load.throbber =
	    gtk_timeout_add(100, (GtkFunction)throbber_loop, browser);
    }

    gtk_widget_set_sensitive(browser->btn.stop, TRUE);
    gtk_widget_set_sensitive(browser->btn.reload, FALSE);
    light_update_status(browser);
}

static void embed_load_finished_cb(GtkMozEmbed *embed, Browser *browser)
{
    browser->load.st_time = 0;
    browser->load.percent = 0;
    browser->load.bytes = 0;
    browser->load.maxbytes = 0;
    
    /* deactivate throbber */
    if (browser->load.throbber) {
	gtk_timeout_remove(browser->load.throbber);
	browser->load.throbber = 0;
	gtk_label_set_text(GTK_LABEL(browser->sb.label), "");
    }
    gtk_widget_set_sensitive(browser->btn.stop, FALSE);
    gtk_widget_set_sensitive(browser->btn.reload, TRUE);
    light_update_status(browser);
}

static int embed_open_uri_cb(GtkMozEmbed *embed, const char *uri,
	Browser *browser)
{
    /* for now */
    return FALSE;
}

static void net_state_change_cb(GtkMozEmbed *embed, int flags,
	unsigned int status, Browser *browser)
{
    if (flags & GTK_MOZ_EMBED_FLAG_IS_REQUEST) {
	if (flags & GTK_MOZ_EMBED_FLAG_REDIRECTING)
	    browser->sb.message = "Redirecting...";
	else if (flags & GTK_MOZ_EMBED_FLAG_TRANSFERRING)
	    browser->sb.message = "Transferring data...";
	else if (flags & GTK_MOZ_EMBED_FLAG_NEGOTIATING)
	    browser->sb.message = "Waiting for authorization...";
    }

    if (status == GTK_MOZ_EMBED_STATUS_FAILED_DNS)
	browser->sb.message = "Site not found";
    else if (status == GTK_MOZ_EMBED_STATUS_FAILED_CONNECT)
	browser->sb.message = "Failed to connect";
    else if (status == GTK_MOZ_EMBED_STATUS_FAILED_TIMEOUT)
	browser->sb.message = "Failed due to connection timeout";
    else if (status == GTK_MOZ_EMBED_STATUS_FAILED_USERCANCELED)
	browser->sb.message = "User canceled connecting";

    if (flags & GTK_MOZ_EMBED_FLAG_IS_DOCUMENT) {
	if (flags & GTK_MOZ_EMBED_FLAG_START)
	    browser->sb.message = "Loading...";
	else if (flags & GTK_MOZ_EMBED_FLAG_STOP)
	    browser->sb.message = "Document: Done";
    }
    light_update_status(browser);
}

static void progress_change_cb(GtkMozEmbed *embed, int cur, int max,
	Browser *browser)
{
    if (max == -1) {
	browser->load.percent = 0;
	browser->load.bytes = cur;
	browser->load.maxbytes = 0;
    } else if (cur > max) {
	browser->load.percent = 100;
	browser->load.bytes = cur;
	browser->load.maxbytes = max;
    } else {
	browser->load.bytes = cur;
	browser->load.maxbytes = max;
	browser->load.percent = (!max) ? 0 : (cur * 100) / max;
    }
    light_update_status(browser);
}

/* dom events */
static int dom_key_press_cb(GtkMozEmbed *embed, void *event, Browser *browser)
{
    KeyEventInfo info;
    unsigned long key;
    int modifier;
    int state = 0;

    moz_key_info(event, &info);
    key = info.key;
    modifier = info.modifier;

    switch (key) {
	case DOM_VK_ESCAPE:
	    gtk_moz_embed_stop_load(embed);
	    return TRUE;
	    break;
	default:
	    break;
    }

    if (modifier & ALT_KEY) {
	switch (key) {
	    case DOM_VK_KP_LEFT:
	    case DOM_VK_LEFT:
		gtk_moz_embed_go_back(embed);
		return TRUE;
		break;
	    case DOM_VK_KP_RIGHT:
	    case DOM_VK_RIGHT:
		gtk_moz_embed_go_forward(embed);
		return TRUE;
		break;
	}
    }

    if (modifier & CTRL_KEY)
	state |= GDK_CONTROL_MASK;
    if (modifier & SHIFT_KEY)
	state |= GDK_SHIFT_MASK;
    if (modifier & ALT_KEY)
	state |= GDK_MOD1_MASK;

    /* We haven't specifically handled the keypress, so send it to the
     * main window in case it is a menu accelerator key */
    if (!gtk_accel_groups_activate(GTK_OBJECT(browser->window),
				   info.key, state))
	return FALSE;

    return TRUE;
}

static int dom_mouse_down_cb(GtkMozEmbed *embed, void *event, Browser *browser)
{
    MouseEventInfo *info;
    int rv;

    /* get the mouse event info */
    info = g_new0(MouseEventInfo, 1);
    rv = moz_mouse_info(event, info, browser);
    g_print("Got %d from moz_mouse_info\n", rv);

    switch (info->button) {
    case 2:
	context_popup_menu(browser, info);
	return TRUE;
	break;
    }

    if (info->linktext)
	g_free(info->linktext);
    g_free(info);
    return FALSE;
}

static int dom_mouse_click_cb(GtkMozEmbed *embed, void *event, Browser *browser)
{
    MouseEventInfo *info;

    /* get the mouse event info */
    info = g_new0(MouseEventInfo, 1);
    moz_mouse_info(event, info, browser);

    /* this needs to be freed */
    switch (info->button) {
	case 0:
	    if ((info->context & CONTEXT_LINK) && (info->modifier == SHIFT_KEY))
		moz_save_link(browser, info->link);
	    break;
	case 1:
	    if (info->context & CONTEXT_LINK)
		light_window_new(info->link);
	    break;
	    return TRUE;
    }

    if (info->linktext)
	g_free(info->linktext);
    g_free(info);
    return FALSE;
}


static void embed_new_orphan_window_cb(GtkMozEmbedSingle *embed,
	GtkMozEmbed **retval, guint chromemask, gpointer data)
{
    Browser *new_browser = NULL;

    new_browser = light_new(chromemask);

    *retval = GTK_MOZ_EMBED(new_browser->embed);
}

/**
 * This is some shit that got added to Mozilla recently to handle weird
 * type of windows
 */
void embed_single_init(void)
{
    GtkMozEmbedSingle *single;

    single = gtk_moz_embed_single_get();

    gtk_signal_connect(GTK_OBJECT(single), "new_window_orphan",
	    GTK_SIGNAL_FUNC(embed_new_orphan_window_cb), NULL);
}
