#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtkmozembed.h>
#include <string.h>
#include "light.h"

#include "pixmaps/back.xpm"
#include "pixmaps/forward.xpm"
#include "pixmaps/reload.xpm"
#include "pixmaps/stop.xpm"
#include "pixmaps/fifo.xpm"

#include "fifo.h"
#include "menu.h"
#include "moz_cb.h"
#include "moz_main.h"
#include "ui.h"
#include "ui_cb.h"
#include "utils.h"

extern int win_count;
extern GList *win_list;
extern Config config;
extern GHashTable *detect;
#ifndef ENGLISH_IS_NO1
extern char **charsets;
#endif
static char *nullicon[] = { "1 1 1 1", "+ c None", "+" };

enum {
    BACK_XPM,
    FIFO_XPM,
    FORWARD_XPM,
    RELOAD_XPM,
    STOP_XPM
};

static int progress_loop(Browser *browser);
static GtkWidget *light_toolbar_new(Browser *browser);
static GtkWidget *light_statbar_new(Browser *browser);
static GtkWidget *light_pixmap(GtkWidget *widget, int type);

static int progress_loop(Browser *browser)
{
    static int val;

    if (!browser->sb.id)
	return FALSE;

    /* update progressive loader "activity" indicator */
    if (++val > 100)
	val = 0;
    
    gtk_progress_set_value(GTK_PROGRESS(browser->sb.progress), val);
    return TRUE;
}

Browser *light_new(unsigned int chrome)
{
    Browser *browser = NULL;
    GtkAccelGroup *accel = NULL;
    GtkWidget *vbox;

    /* new browser setup */
    browser = g_new0(Browser, 1);
    browser->size = 100;
    browser->init = 1;
    win_list = g_list_prepend(win_list, browser);
    win_count++;

    /* setup chrome */
    if (chrome == GTK_MOZ_EMBED_FLAG_DEFAULTCHROME)
	chrome = GTK_MOZ_EMBED_FLAG_ALLCHROME;

    /* keyboard accelerator group */
    accel = gtk_accel_group_new();

    /* create main browser window */
    browser->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size(GTK_WINDOW(browser->window), config.width,
	    config.height);
    gtk_window_set_policy(GTK_WINDOW(browser->window), TRUE, TRUE, TRUE);
    gtk_window_set_title(GTK_WINDOW(browser->window), "Light");
    light_selection_setup(browser);
    gtk_signal_connect(GTK_OBJECT(browser->window), "destroy",
	    GTK_SIGNAL_FUNC(mainwindow_destroy_cb), browser);
   
    /* all the stuff gets packed here */
    vbox = gtk_vbox_new(FALSE, NO_SPACING);
    gtk_widget_show(vbox);
    gtk_container_add(GTK_CONTAINER(browser->window), vbox);

    /* menubar */
    browser->menubar = light_menubar_new(accel, browser);
    gtk_menu_bar_set_shadow_type(GTK_MENU_BAR(browser->menubar),
	    GTK_SHADOW_ETCHED_IN);
    gtk_box_pack_start(GTK_BOX(vbox), browser->menubar, FALSE, FALSE,
	    NO_SPACING);

    /* toolbar and url entry */
    browser->toolbar = light_toolbar_new(browser);
    gtk_box_pack_start(GTK_BOX(vbox), browser->toolbar, FALSE, FALSE,
	    NO_SPACING);

    /* gtkmozembed */
    browser->embed = gtk_moz_embed_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->embed, TRUE, TRUE,
	    NO_SPACING);
    gtk_widget_set_usize(browser->embed, 400, 400);

    /* statusbar */
    browser->statbar = light_statbar_new(browser);
    gtk_box_pack_start(GTK_BOX(vbox), browser->statbar, FALSE, FALSE,
	    NO_SPACING);

    /* connect all gtkmozembed signals */
    moz_connect_signals(browser);

    gtk_moz_embed_set_chrome_mask(GTK_MOZ_EMBED(browser->embed),
	    chrome);

    gtk_window_add_accel_group(GTK_WINDOW(browser->window), accel);

    /* prevent idiots from changing shortcut keys */
    gtk_accel_group_lock(accel);

#ifndef ENGLISH_IS_NO1
    /* now, get the charsets if we haven't already, and populate the
     * charset menu */
    if (charsets == NULL)
	moz_get_charset(&charsets);

    /* this assumes charsets[] has something in it */
    fill_charset_menu(browser);
#endif

    /* select the autodetect menu */
    g_hash_table_foreach(detect, light_set_detect, browser);

    /* initialize complete */
    browser->init = 0;
    gtk_widget_grab_focus(browser->entry);

    return browser;
}

static GtkWidget *light_toolbar_new(Browser *browser)
{
    GtkWidget *handle;
    GtkToolbar *toolbar;

    /* handle = gtk_handle_box_new(); frame looks better? */
    handle = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(handle), GTK_SHADOW_ETCHED_IN);
    toolbar = GTK_TOOLBAR(gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL,
		GTK_TOOLBAR_BOTH));
    gtk_container_add(GTK_CONTAINER(handle), GTK_WIDGET(toolbar));

    gtk_toolbar_set_space_size(toolbar, UI_SPACING);
    gtk_container_set_border_width(GTK_CONTAINER(toolbar), UI_SPACING);
    gtk_toolbar_set_button_relief(toolbar, GTK_RELIEF_NONE);

    browser->btn.back = gtk_toolbar_append_item(toolbar,
	    config.pixmaps != 1 ? "Back" : NULL,
	    "Go Back (Alt+Left Arrow)",
	    NULL,
	    light_pixmap(GTK_WIDGET(toolbar), BACK_XPM),
	    back_cb, /* callback */
	    browser);
    gtk_toolbar_append_space(toolbar);

    browser->btn.forward = gtk_toolbar_append_item(toolbar,
	    config.pixmaps != 1 ? "Forward" : NULL,
	    "Go Back (Alt+Left Arrow)",
	    NULL,
	    light_pixmap(GTK_WIDGET(toolbar), FORWARD_XPM),
	    forward_cb, /* callback */
	    browser);
    gtk_toolbar_append_space(toolbar);

    browser->btn.reload = gtk_toolbar_append_item(toolbar,
	    config.pixmaps != 1 ? "Reload" : NULL,
	    "Reload Current Page",
	    NULL,
	    light_pixmap(GTK_WIDGET(toolbar), RELOAD_XPM),
	    reload_cb, /* callback */
	    browser);
    gtk_toolbar_append_space(toolbar);

    browser->btn.stop = gtk_toolbar_append_item(toolbar,
	    config.pixmaps != 1 ? "Stop" : NULL,
	    "Stop (Escape)",
	    NULL,
	    light_pixmap(GTK_WIDGET(toolbar), STOP_XPM),
	    stop_cb, /* callback */
	    browser);
    gtk_toolbar_append_space(toolbar);

    /* URL entry box */
    browser->entry = gtk_entry_new();
    gtk_toolbar_append_widget(toolbar, browser->entry,
	    "Enter a URL here", NULL);
    gtk_signal_connect(GTK_OBJECT(browser->entry), "button_press_event",
	    GTK_SIGNAL_FUNC(url_click_cb), browser);
    gtk_toolbar_append_space(toolbar);
    /* hook up the activate signal for the url entry */
    gtk_signal_connect(GTK_OBJECT(browser->entry), "activate",
		       GTK_SIGNAL_FUNC(url_activate_cb), browser);
    /* hook right click signal to display a menu there */
    gtk_signal_connect(GTK_OBJECT(browser->entry), "button_press_event",
	    GTK_SIGNAL_FUNC(url_rightclick_cb), browser);
    /* hook up keypress event to catch "esc" */
    gtk_signal_connect(GTK_OBJECT(browser->entry), "key_press_event",
	    GTK_SIGNAL_FUNC(url_keypress_cb), browser);

    browser->btn.fifo = gtk_toolbar_append_item(toolbar,
	    config.pixmaps != 1 ? "Fifo" : NULL,
	    "Bookmarks (Alt+I)",
	    NULL,
	    light_pixmap(GTK_WIDGET(toolbar), FIFO_XPM),
	    NULL, /* callback */
	    browser);
    gtk_toolbar_append_space(toolbar);
    gtk_signal_connect(GTK_OBJECT(browser->btn.fifo), "button_press_event",
	    GTK_SIGNAL_FUNC(fifo_pressed_cb), browser);

    gtk_widget_set_sensitive(browser->btn.back, FALSE);
    gtk_widget_set_sensitive(browser->btn.forward, FALSE);
    gtk_widget_set_sensitive(browser->btn.reload, FALSE);
    gtk_widget_set_sensitive(browser->btn.stop, FALSE);

    /* hook entry resize event */
    gtk_signal_connect(GTK_OBJECT(toolbar), "size_allocate",
		       GTK_SIGNAL_FUNC(toolbar_resize_cb), browser);
    return handle;
}

static GtkWidget *light_statbar_new(Browser *browser)
{
    GtkWidget *hbox;

    hbox = gtk_hbox_new(FALSE, UI_SPACING);

    /* throbber */
    browser->sb.frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(browser->sb.frame), GTK_SHADOW_IN);
    gtk_widget_set_usize(browser->sb.frame, 16, -1);

    browser->sb.label = gtk_label_new("");
    gtk_misc_set_alignment(GTK_MISC(browser->sb.label), 0.5, 0.5);
    gtk_container_add(GTK_CONTAINER(browser->sb.frame), browser->sb.label);
    gtk_box_pack_start(GTK_BOX(hbox), browser->sb.frame, FALSE, TRUE,
	    NO_SPACING);
    gtk_widget_show_all(browser->sb.frame);

    /* statusbar */
    browser->sb.bar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(hbox), browser->sb.bar, TRUE, TRUE,
	    NO_SPACING);
    gtk_widget_show(browser->sb.bar);

    /* progress bar */
    browser->sb.progress = gtk_progress_bar_new();
    gtk_box_pack_start(GTK_BOX(hbox), browser->sb.progress, FALSE, TRUE,
	    NO_SPACING);
    gtk_widget_show(browser->sb.progress);

    return hbox;
}

void light_window_new(const char *url)
{
    Browser *browser = light_new(GTK_MOZ_EMBED_FLAG_DEFAULTCHROME);
    char *stream = "<html><body></body></html>";

    light_set_visibility(browser, TRUE);

    /* load up a url */
    if (url) {
	if (!g_strcasecmp(url, "about:blank"))
	    gtk_moz_embed_render_data(GTK_MOZ_EMBED(browser->embed),
		    stream, strlen(stream), "file://", "text/html");
	else
	    gtk_moz_embed_load_url(GTK_MOZ_EMBED(browser->embed), url);
    }
}

void light_find(Browser *browser)
{
    GtkWidget *dialog;
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *label;
    GtkWidget *entry;
    GtkWidget *sens, *back;
    GtkWidget *button;
    GtkAccelGroup *accel;
    int key;

    dialog = gtk_dialog_new();
    accel = gtk_accel_group_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "Find");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, TRUE);
    gtk_object_set_data(GTK_OBJECT(dialog), "browser", browser);

    /* main vbox */
    vbox = gtk_vbox_new(FALSE, UI_SPACING);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), UI_SPACING);
    /* put it inside the dialog */
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox, TRUE, TRUE,
		       0);

    /* Find: label */
    label = gtk_label_new("Find:");
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);

    /* find entry - doubleclickable */
    entry = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(vbox), entry, TRUE, TRUE, 0);
    gtk_object_set_data(GTK_OBJECT(dialog), "entry", entry);
    gtk_signal_connect(GTK_OBJECT(entry), "activate",
		       GTK_SIGNAL_FUNC(find_dialog_find_cb), dialog);
    gtk_signal_connect(GTK_OBJECT(entry), "key_press_event",
		       GTK_SIGNAL_FUNC(dialog_keypress_cb), dialog);

    /* if we had some previous searches, set that text in, and select it
     * so that the user can quickly type over, if needed */
    if (browser->find) {
	gtk_entry_set_text(GTK_ENTRY(entry), browser->find);
	gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);
    }

    /* modifier checkboxes */
    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

    sens = gtk_check_button_new_with_label("");
    key = gtk_label_parse_uline(GTK_LABEL(GTK_BIN(sens)->child),
	    "Case _Sensitive");
    gtk_object_set_data(GTK_OBJECT(dialog), "sens", sens);
    gtk_box_pack_start(GTK_BOX(hbox), sens, FALSE, FALSE, 0);
    gtk_signal_connect_object(GTK_OBJECT(sens), "clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_grab_focus),
			      (gpointer) entry);
    gtk_widget_add_accelerator(sens, "clicked", accel,
                              key, GDK_MOD1_MASK, 0);

    back = gtk_check_button_new_with_label("");
    key = gtk_label_parse_uline(GTK_LABEL(GTK_BIN(back)->child),
	    "Find Back_wards");
    gtk_object_set_data(GTK_OBJECT(dialog), "back", back);
    gtk_box_pack_start(GTK_BOX(hbox), back, FALSE, FALSE, 0);
    gtk_signal_connect_object(GTK_OBJECT(back), "clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_grab_focus),
			      (gpointer) entry);
    gtk_widget_add_accelerator(back, "clicked", accel,
                              key, GDK_MOD1_MASK, 0);

    /* bottom part */
    gtk_container_set_border_width(GTK_CONTAINER
				   (GTK_DIALOG(dialog)->action_area),
				   UI_SPACING);
    hbox = gtk_hbox_new(TRUE, UI_SPACING);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), hbox,
		       TRUE, TRUE, 0);

    button = gtk_button_new_with_label("Find");
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(find_dialog_find_cb), dialog);

    button = gtk_button_new_with_label("");
    key = gtk_label_parse_uline(GTK_LABEL(GTK_BIN(button)->child),
	    "Cl_ear");
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(find_dialog_clear_cb), dialog);
    gtk_widget_add_accelerator(button, "clicked", accel,
                              key, GDK_MOD1_MASK, 0);

    button = gtk_button_new_with_label("Close");
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(find_dialog_close_cb), dialog);

    gtk_widget_show_all(dialog);
    gtk_widget_grab_focus(entry);
    gtk_window_add_accel_group(GTK_WINDOW(dialog), accel);
}

/* searchbox */
void light_searchbox_new(Browser *browser)
{
    GtkWidget *window;
    GtkWidget *label;
    GtkWidget *entry;
    GtkWidget *table;
    GtkAccelGroup *accel;
    int key;

    window = gtk_window_new(GTK_WINDOW_DIALOG);
    accel = gtk_accel_group_new();

    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
    gtk_window_set_title(GTK_WINDOW(window), "Search");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, TRUE);
    gtk_widget_realize(window);
    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
		       GTK_SIGNAL_FUNC(dialog_keypress_cb), window);

    table = gtk_table_new(3, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(window), UI_SPACING);
    gtk_container_add(GTK_CONTAINER(window), table);

    /* google search */
    entry = gtk_entry_new();
    label = gtk_label_new("");
    key = gtk_label_parse_uline(GTK_LABEL(label), "_Google: ");
    gtk_widget_add_accelerator(entry, "grab_focus", accel, key, GDK_MOD1_MASK, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
    gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 0, 1);
    gtk_object_set_data(GTK_OBJECT(entry), "search_type", GINT_TO_POINTER(10));
    gtk_object_set_data(GTK_OBJECT(entry), "window", window);
    gtk_signal_connect(GTK_OBJECT(entry), "activate",
	    GTK_SIGNAL_FUNC(search_activate_cb), browser);
    gtk_widget_grab_focus(entry);

    /* freshmeat search */
    entry = gtk_entry_new();
    label = gtk_label_new("");
    key = gtk_label_parse_uline(GTK_LABEL(label), "Fresh_Meat: ");
    gtk_widget_add_accelerator(entry, "grab_focus", accel, key, GDK_MOD1_MASK, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 1, 2);
    gtk_object_set_data(GTK_OBJECT(entry), "search_type", GINT_TO_POINTER(20));
    gtk_object_set_data(GTK_OBJECT(entry), "window", window);
    gtk_signal_connect(GTK_OBJECT(entry), "activate",
	    GTK_SIGNAL_FUNC(search_activate_cb), browser);

    /* user-specified search */
    entry = gtk_entry_new();
    label = gtk_label_new("");
    key = gtk_label_parse_uline(GTK_LABEL(label), "C_ustom: ");
    gtk_widget_add_accelerator(entry, "grab_focus", accel, key, GDK_MOD1_MASK, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 2, 3);
    gtk_table_attach_defaults(GTK_TABLE(table), entry, 1, 2, 2, 3);
    gtk_object_set_data(GTK_OBJECT(entry), "search_type", GINT_TO_POINTER(30));
    gtk_object_set_data(GTK_OBJECT(entry), "window", window);
    gtk_signal_connect(GTK_OBJECT(entry), "activate",
	    GTK_SIGNAL_FUNC(search_activate_cb), browser);

    gtk_widget_show_all(window);
    gtk_window_add_accel_group(GTK_WINDOW(window), accel);
}

/* fifo */
void fifo_keypress_cb(GtkWidget *widget, GdkEventKey *key, Browser *browser)
{
    GtkCList *clist;
    GtkWidget *window;
    int row = 0;
    char *url = NULL;

    clist = gtk_object_get_data(GTK_OBJECT(widget), "clist");
    window = gtk_object_get_data(GTK_OBJECT(clist), "window");

    switch (key->keyval) {
	case GDK_Escape:
	    gtk_widget_destroy(window);
	    break;
	case GDK_space:
	case GDK_KP_Enter:
	case GDK_ISO_Enter:
	case GDK_Return:
	    row = clist->focus_row;
	    switch (key->state) {
		case GDK_CONTROL_MASK:
		    fifo_nuke(row);
		    gtk_widget_destroy(window);
		    break;
		case GDK_MOD1_MASK:
		    gtk_clist_get_text(clist, row, 2, &url);
		    light_copy_clipboard(browser, url);
		    gtk_widget_destroy(window);
		    break;
		default:
		    gtk_clist_get_text(clist, row, 2, &url);
		    gtk_moz_embed_load_url(GTK_MOZ_EMBED(browser->embed), url);
		    if (key->state == GDK_SHIFT_MASK)
			fifo_nuke(row);
		    gtk_widget_destroy(window);
		    break;
	    }
	default:
	    break;
    }
}

static void fifo_click_cb(GtkWidget *widget, GdkEventButton *event, Browser *browser)
{
    GtkCList *clist;
    GtkWidget *window;
    int row = 0;
    char *url = NULL;

    /* this is generated by "wheel scrolling" of the clist, so we ignore it */
    if (event->button > 3)
	return;

    clist = gtk_object_get_data(GTK_OBJECT(widget), "clist");
    window = gtk_object_get_data(GTK_OBJECT(clist), "window");
    row = clist->focus_row;

    /* control-click just deletes the fifo entry */
    if (event->state & GDK_CONTROL_MASK) {
	fifo_nuke(row);
	gtk_widget_destroy(window);
	return;
    }

    /* alt-click copies URL to the clipboard */
    if (event->state & GDK_MOD1_MASK) {
	gtk_clist_get_text(clist, row, 2, &url);
	light_copy_clipboard(browser, url);
	gtk_widget_destroy(window);
	return;
    }

    /* double-click loads it */
    if (event->type == GDK_2BUTTON_PRESS) {
	gtk_clist_get_text(clist, row, 2, &url);
	gtk_moz_embed_load_url(GTK_MOZ_EMBED(browser->embed), url);

	/* shift-click loads the url (above) and deletes it from fifo */
	if (event->state & GDK_SHIFT_MASK)
	    fifo_nuke(row);

	gtk_widget_destroy(window);
    }
}

GtkWidget *light_fifo_new(Browser *browser)
{
    GtkWidget *window;
    GtkWidget *clist;
    GtkWidget *scroll;

    char *titles[] = { "ID", "Name", "URL" };

    window = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, TRUE);
    gtk_widget_realize(window);

    /* handle fifo list keyboard navigation */
    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
		       GTK_SIGNAL_FUNC(fifo_keypress_cb), browser);
    /* handle clicking on the list */
    gtk_signal_connect(GTK_OBJECT(window), "button_press_event",
	    GTK_SIGNAL_FUNC(fifo_click_cb), browser);

    /* this is a hack, move it off the screen so that it doesnt flicker */
    gdk_window_move(GTK_WIDGET(window)->window, -4096, 0);
    
    /* make clist */
    clist = gtk_clist_new_with_titles(3, titles);
    /* uhhh */
    gtk_clist_set_reorderable(GTK_CLIST(clist), FALSE);
    gtk_clist_set_use_drag_icons(GTK_CLIST(clist), FALSE);
    gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_BROWSE);
    gtk_clist_column_titles_passive(GTK_CLIST(clist));
    
    gtk_clist_set_column_width(GTK_CLIST(clist), 0, 24);
    gtk_clist_set_column_width(GTK_CLIST(clist), 1, 128);
    gtk_clist_set_column_width(GTK_CLIST(clist), 2, 512);
    gtk_clist_set_column_resizeable(GTK_CLIST(clist), 0, FALSE);
    gtk_clist_set_column_resizeable(GTK_CLIST(clist), 1, FALSE);
    gtk_clist_set_column_resizeable(GTK_CLIST(clist), 2, FALSE);

    gtk_object_set_data(GTK_OBJECT(window), "clist", clist);
    gtk_object_set_data(GTK_OBJECT(clist), "window", window);

    /* fill clist with stuff */
    fifo_list(clist);

    /* make scrolled window */
    scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
	    GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(window), scroll);
    gtk_container_add(GTK_CONTAINER(scroll), clist);

    /* resize the scrolled window to perfectly fit in the clist */
    gtk_widget_set_usize(GTK_WIDGET(scroll), clist->requisition.width,
	    (clist->requisition.height > config.fifoheight) ?
	    config.fifoheight : clist->requisition.height);

    gtk_widget_show_all(window);

    return window;
}

static GtkWidget *light_pixmap(GtkWidget *widget, int type)
{
    GdkColormap *colormap;
    GdkPixmap *gdkpixmap = NULL;
    GdkBitmap *mask;
    GtkWidget *pixmap;
    char *pixdir[] = { "pixmaps/back.xpm", "pixmaps/fifo.xpm",
		       "pixmaps/forward.xpm", "pixmaps/reload.xpm",
		       "pixmaps/stop.xpm" };
    char **pixloc[] = { back_xpm, fifo_xpm, forward_xpm, reload_xpm, stop_xpm };
    char *file = NULL;

    colormap = gtk_widget_get_colormap(widget);

    if (config.pixmaps > 0) {
	/* make pixmaps */
	file = check_file_exists(config.pref, pixdir[type]);
	if (file) {
	    gdkpixmap = gdk_pixmap_colormap_create_from_xpm(NULL, colormap,
		    &mask, NULL, file);
	} else {
	    gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, colormap,
		    &mask, NULL, pixloc[type]);
	}
    } else {
	gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, colormap,
		&mask, NULL, nullicon);
    }

    if (gdkpixmap == NULL)
	return NULL;

    pixmap = gtk_pixmap_new(gdkpixmap, mask);
    gdk_pixmap_unref(gdkpixmap);
    gdk_bitmap_unref(mask);
    if (file)
	g_free(file);
    return pixmap;
}

void light_set_visibility(Browser *browser, int vis)
{
    unsigned int chrome;

    if (!vis) {
	gtk_widget_hide(browser->window);
	return;
    }

    chrome = gtk_moz_embed_get_chrome_mask(GTK_MOZ_EMBED(browser->embed));

    if (chrome == GTK_MOZ_EMBED_FLAG_DEFAULTCHROME)
	chrome = GTK_MOZ_EMBED_FLAG_ALLCHROME;

    if (chrome & GTK_MOZ_EMBED_FLAG_MENUBARON)
	gtk_widget_show_all(browser->menubar);
    else
	gtk_widget_hide_all(browser->menubar);

    if (chrome & GTK_MOZ_EMBED_FLAG_TOOLBARON)
	gtk_widget_show_all(browser->toolbar);
    else
	gtk_widget_hide_all(browser->toolbar);

    if (chrome & GTK_MOZ_EMBED_FLAG_STATUSBARON)
	gtk_widget_show_all(browser->statbar);
    else
	gtk_widget_hide_all(browser->statbar);

    gtk_widget_show(browser->embed);
    if (!GTK_WIDGET_VISIBLE(browser->window))
	gtk_widget_show(browser->window);
}

void light_update_status(Browser *browser)
{
    char message[256];
    char *status = NULL;

    /* this is kind of a cut & paste from galeon :/ */
    /* handle progress bar */
    if (browser->load.bytes > 0) {
	if (browser->load.percent == 0) {
	    gtk_progress_set_activity_mode(GTK_PROGRESS(browser->sb.progress), TRUE);
	    if (!browser->sb.id) {
		browser->sb.id = gtk_timeout_add(100, (GtkFunction)progress_loop, browser);
	    }
	} else {
	    browser->sb.id = 0;
	    gtk_progress_set_activity_mode(GTK_PROGRESS(browser->sb.progress),
		    FALSE);
	    gtk_progress_set_value(GTK_PROGRESS(browser->sb.progress),
		    browser->load.percent);
	}
    } else {
	browser->sb.id = 0;
	gtk_progress_set_activity_mode(GTK_PROGRESS(browser->sb.progress), FALSE);
	gtk_progress_set_value(GTK_PROGRESS(browser->sb.progress), 0);
    }

    /* handle status bar */
    gtk_statusbar_pop(GTK_STATUSBAR(browser->sb.bar), 1);

    /* get link message going */
    if (browser->sb.link) {
	gtk_statusbar_push(GTK_STATUSBAR(browser->sb.bar), 1,
			   browser->sb.link);
	return;
    }

    status = (browser->sb.message) ? browser->sb.message : "";

    if (browser->load.bytes == 0) {
	g_snprintf(message, 255, "%s", status);
    } else if (browser->load.bytes <= browser->load.maxbytes) {
	char time_remaining[64] = "";
	int hours, minutes, seconds;
	time_t time_delta;

	time_delta = (time(NULL) - browser->load.st_time);
	if ((browser->load.percent > 5) &&
		(browser->load.percent < 100) && time_delta > 0) {
	    time_delta *= ((double)browser->load.maxbytes /
		    (double)browser->load.bytes) - 1.0;
	    hours = time_delta / (60 * 60);
	    time_delta %= (60 * 60);
	    minutes = time_delta / 60;
	    seconds = time_delta % 60;
	    g_snprintf(time_remaining, 64,
		    ", %02d:%02d:%02d remaining",
		    hours, minutes, seconds);
	}
	g_snprintf(message, 255,
		"%s (%d%% complete, %d kB of %d kB loaded%s)",
		status,
		browser->load.percent, browser->load.bytes >> 10,
		browser->load.maxbytes >> 10,
		time_remaining);
    } else {
	g_snprintf(message, 255, "%s (%d kB loaded)",
		status, browser->load.bytes >> 10);
    }

    gtk_statusbar_push(GTK_STATUSBAR(browser->sb.bar), 1, message);
}
